PSEUDOTASK = {
    "morgan": (1024, "classification"),
    "maccs": (167, "classification"), 
    "logP": (1, "regression"),
    # "SAS": (1, "regression"),
    # "TPSA": (1, "regression"),
    # "QED": (1, "regression"),
    # "cycle_score": (1, "regression"),
    # "SA_score": (1, "regression"),
    # "weight": (1, "regression"),
}